%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%											
%   "Copyright (c) 2011 David I Shuman"				
%   All rights reserved.
%
%   This is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This code is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   If you do not have a copy of the GNU General Public License,
%   see <http://www.gnu.org/licenses/>.									
%											
%   Author: David I Shuman
%   Signal Processing Laboratories LTS2 and LTS4
%   cole Polytechnique Fdrale de Lausanne (EPFL)
%   Contact: david.shuman@epfl.ch
%
%   Date last modified:	10/12/2011 
% 
%   Description:
%   
%   Runs the weighted iterative soft thresholding algorithm to solve the 
%   weighted LASSO:
%   min_a ||y-W*a||_2^2 + lambda*||a||_{1,v} 
%   where y=x+noise
%
%   Inputs:
%   1. parameters, which includes parameters.lambda, parameters.step (the
%   step size in each thresholding iteration, and parameters.num_iterations,
%   the number of thresholding iterations
%   2. weights, which corresponds to the weights v in the above problem
%   3. The corrupted signal y
%   4. The number of wavelets; i.e., the number of rows in W
%   5. The synthesis operator W* in operator form
%   6. The analysis operator W in operator form
%
%   Outputs:
%   1. The estimated unmasked signal x_hat  
%   2. The reconstructed wavelet coefficients; i.e., the solution a^* to the
%   above optimization problem
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [reconstruction, reconstructed_coefficients]=WISTA2(parameters, weights, corrupted_signal, num_wavelets, Synthesis_Op, Analysis_Op)
lambda = parameters.lambda;
step = parameters.step;
num_iterations = parameters.num_iterations;

Coeff_new=zeros([num_wavelets,1]);
measured_coeffs=Analysis_Op(corrupted_signal);
for i = 1:num_iterations
    Coeff_old = Coeff_new;
    updated_coefficients = Coeff_old+step*measured_coeffs-step*Analysis_Op(Synthesis_Op(Coeff_old));
    Coeff_new=sign(updated_coefficients).*max(abs(updated_coefficients)-lambda*step*weights, 0);
end
reconstructed_coefficients = Coeff_new;
reconstruction = Synthesis_Op(Coeff_new);
