%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%											
%   "Copyright (c) 2011 David I Shuman"				
%   All rights reserved.
%
%   This is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This code is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   If you do not have a copy of the GNU General Public License,
%   see <http://www.gnu.org/licenses/>.									
%											
%   Author: David I Shuman
%   Signal Processing Laboratories LTS2 and LTS4
%   cole Polytechnique Fdrale de Lausanne (EPFL)
%   Contact: david.shuman@epfl.ch
%
%   Date last modified:	10/12/2011 
% 
%   This file requires the Spectral Graph Wavelets (SGWT) Toolbox, which is
%   is written by David K. Hammond, and is available at
%   http://wiki.epfl.ch/sgwt
% 
%   It also requires the functions plot_graph_signal.m and show_network.m,
%   as well as the MatlabBGL library, which is available at
%   http://www.mathworks.com/matlabcentral/fileexchange/10922
%
%   Description:
%     This file executes the distributed denoising experiment in Section
%     VI.A of the paper "Distributed Signal Processing via Chebyshev
%     Polynomial Approximation," which has been submitted to the IEEE 
%     Transactions on Signal Processing. It solves the distributed Tikhonov 
%     regularization problem in equation (16). Figures 2-5 of the paper, as 
%     well as the numerical results in Section VI.A, are generated from 
%     this code. Note that the code implements the computations in a 
%     centralized manner. 
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
close all;
randn('seed', 18); rand('seed', 18);

num_iterations = 1000;
noise_MSEs = zeros(num_iterations,1);
reconstruction_MSEs = zeros(num_iterations,1);
noise_SNRs = zeros(num_iterations,1);
reconstruction_SNRs = zeros(num_iterations,1);
connecteds = zeros(num_iterations,1);

% Parameters of the random graphs
N = 500;
plane_dim = 1;
sigma=.5; 
target_dist_cutoff = .075;
thresh = .6; 
s = sqrt(-target_dist_cutoff^2/(2*log(thresh)));
distance_cutoff = sqrt(-2*s^2*log(thresh));

% Parameters of the approximate multiplier
tau=1;
smoothing_filter = @(x) tau./(tau+2*x);
K = 15; % Chebyshev approximation order

% Solve the regularization problem:
% min_f f'Lf + tau/2 * norm(f-fo)_2^2
% by applying the filter tau/(tau+2lambda_l)

% Each iteration is a repetition of the experiment
for iter=1:num_iterations

    % Generate data on a plane
    XCoords = plane_dim*rand(N,1);
    YCoords = plane_dim*rand(N,1);
    
    % Generate the true signal
    fo = (XCoords.^2)+YCoords.^2-1; 

    % Corrupt the signal with noise
    noise = randn(N,1)*sigma;
    noisy_signal = fo + noise;

    % Compute distance-based Laplacian 
    d = distanz([XCoords,YCoords]'); 
    A = exp(-d.^2/(2*s^2)); 
    A(A<thresh) = 0; % Thresholding to have a sparse adjacency matrix
    L = sgwt_laplacian(A);

    % Check graph connectivity
    connecteds(iter)=sgwt_check_connected(L);

    % Compute Chebyshev coefficients of the multiplier
    lmax=sgwt_rough_lmax(L);
    arange=[0,lmax];
    cheb_coeffs=sgwt_cheby_coeff(smoothing_filter,K,K+1,arange);

    % Perform the approximate filtering
    smoothed_signal =sgwt_cheby_op(noisy_signal,L,cheb_coeffs,arange);

    % Compute mean-square errors
    noise_MSEs(iter) = mean((noisy_signal-fo).^2);
    reconstruction_MSEs(iter)= mean((smoothed_signal-fo).^2);

    % Compute SNRs
    noise_SNRs(iter)=snr(fo,noisy_signal);
    reconstruction_SNRs(iter)=snr(fo,smoothed_signal);
    
    % Save some information from the first iteration for the figure plots
    if iter == 1
        firstL=L;
        first_fo=fo;
        first_XCoords=XCoords;
        first_YCoords=YCoords;
        first_noise=noise;
        first_noisy_signal=noisy_signal;
        first_smoothed_signal=smoothed_signal;
        first_lmax=lmax;
    end
end

% Output of numerical experiments
fprintf('The mean MSE of all noisy signals is %f.\n',mean(noise_MSEs))
fprintf('The mean MSE of all reconstructed signals is %f.\n',mean(reconstruction_MSEs))
fprintf('%d out of %d graphs are connected.\n',sum(connecteds),num_iterations)
fprintf('The mean MSE of noisy signals on connected graphs is %f.\n',mean(noise_MSEs(connecteds==1)))
fprintf('The mean MSE of reconstructed signals on connected graphs is %f.\n',mean(reconstruction_MSEs(connecteds==1)))

% Figure 2 - Network structure and fo
delta.x=.01;
delta.y=.01;
data = zeros(1+plane_dim/delta.x,1+plane_dim/delta.y);
for i=1:1+plane_dim/delta.x
    for j=1:1+plane_dim/delta.y
        x_i=(i-1)*delta.x;
        y_j=(j-1)*delta.y;
        data(i,j)=(x_i)^2+(y_j)^2-1;
    end
end
adj_matrix = -sign(full(firstL));
adj_matrix = adj_matrix-diag(diag(adj_matrix));
show_network(data,adj_matrix,[first_XCoords,first_YCoords],N,plane_dim,delta);

% Figure 3 - Eigenvectors of the graph Laplacian
[V,E]=eig(full(firstL));
eig_vals = diag(E);
[sorted_eig_vals, indices] = sort(eig_vals,'ascend');

limits.xlow=0;
limits.xhigh=plane_dim; 
limits.ylow=0; 
limits.yhigh=plane_dim; 

z_scale_factor=1.1;
z_vals=V(:,[indices(1),indices(2),indices(3),indices(51)]);
z_abs=max(max(abs(z_vals)));
limits.zlow=-z_scale_factor*z_abs;
limits.zhigh=z_scale_factor*z_abs;

plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,V(:,indices(1)),limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,V(:,indices(2)),limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,V(:,indices(3)),limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,V(:,indices(51)),limits)

% Figure 4 - Chebyshev polynomial approximations to the multiplier
upper=ceil(first_lmax);
lambda=0:.1:upper;
K_vals = [5 15]; 
first_cheb_coeffs = cell(length(K_vals),1);
for i=1:length(K_vals)
    first_cheb_coeffs{i}=sgwt_cheby_coeff(smoothing_filter,K_vals(i),K_vals(i)+1,[0,upper]);
end

figure;
hold on;
plot(lambda,tau./(tau+2*lambda),'k','LineWidth',3); 
plot(lambda,sgwt_cheby_eval(lambda,first_cheb_coeffs{1},[0,upper]),'r','LineWidth',1);
plot(lambda,sgwt_cheby_eval(lambda,first_cheb_coeffs{2},[0,upper]),'g','LineWidth',1);
legend('Exact Multiplier','Chebyshev Polynomial Approximation, K=5','Chebyshev Polynomial Approximation, K=15');
legend boxoff
xlabel('\lambda')

% Figure 5 - Denoisoing figures
z_scale_factor=1.1;
z_vals=[first_fo,first_noise,first_noisy_signal,first_smoothed_signal];
z_abs=max(max(abs(z_vals)));
limits.zlow=-z_scale_factor*z_abs;
limits.zhigh=z_scale_factor*z_abs;

plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,first_fo,limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,first_noise,limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,first_noisy_signal,limits)
plot_graph_signal(first_XCoords,first_YCoords,adj_matrix,N,first_smoothed_signal,limits)



