%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%											
%   "Copyright (c) 2011 David I Shuman"				
%   All rights reserved.
%
%   This is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This code is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   If you do not have a copy of the GNU General Public License,
%   see <http://www.gnu.org/licenses/>.									
%											
%   Author: David I Shuman
%   Signal Processing Laboratories LTS2 and LTS4
%   cole Polytechnique Fdrale de Lausanne (EPFL)
%   Contact: david.shuman@epfl.ch
%
%   Date last modified:	10/12/2011 
% 
%   This file requires the Spectral Graph Wavelets (SGWT) Toolbox, which is
%   is written by David K. Hammond, and is available at
%   http://wiki.epfl.ch/sgwt
%
%   It also requires the functions exact_forward2.m, sgwt_adjoint2.m,
%   sgwt_adjoint_exact2.m, and WISTA2.m, as well as the MatlabBGL library, 
%   which is available at
%   http://www.mathworks.com/matlabcentral/fileexchange/10922
%
%   Description:
%     This file executes the distributed denoising experiment in Section
%     VI.B of the paper "Distributed Signal Processing via Chebyshev
%     Polynomial Approximation," which has been submitted to the IEEE 
%     Transactions on Signal Processing. It solves the distributed LASSO 
%     regularization problem in equations (17) and (19). The numerical 
%     results in Section VI.B are also generated from this code. Note that 
%     the code implements the computations in a centralized manner.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
close all;
randn('seed', 18); rand('seed', 18);

num_iterations = 1000;
noise_MSEs = zeros(num_iterations,1);
reconstruction_MSEs = zeros(num_iterations,1);
reconstruction_approx_wavelet_MSEs=zeros(num_iterations,1);
reconstruction_exact_wavelet_MSEs=zeros(num_iterations,1);
connecteds = zeros(num_iterations,1);

% Parameters of the random graphs
N = 500;
plane_dim = 1;
sigma=.5; 
target_dist_cutoff = .075;
thresh = .6; 
s = sqrt(-target_dist_cutoff^2/(2*log(thresh)));
distance_cutoff = sqrt(-2*s^2*log(thresh));

% Parameters of the approximate multiplier for Tikhonov regularization
tau=1;
smoothing_filter = @(x) tau./(tau+2*x);
K = 15; % Chebyshev approximation order

% Parameters for distributed wavelet denoising
wavelet_weight=.75; 
num_scales=6;
num_wavelets = N*(num_scales+1);
parameters.lambda=1; 
parameters.step=.2; % .25;
parameters.num_iterations = 300; %150;
weights = wavelet_weight*ones([num_wavelets,1]);
weights(1:N)=.01; % Do not promote sparsity in the scaling coefficients
   
% Each iteration is a repetition of the experiment
for iter=1:num_iterations
    if mod(iter,100)==0
        fprintf('Iteration %f.\n',iter)
    end
    
    % Generate data on a plane
    XCoords = plane_dim*rand(N,1);
    YCoords = plane_dim*rand(N,1);
    
    % Generate the true signal
    fo_a = -2*XCoords+.5; 
    fo_b = (XCoords.^2)+YCoords.^2+.5;
    split=1-XCoords < YCoords;
    fo=split.*fo_b+(1-split).*fo_a;
    
    % Corrupt the signal with noise
    noise = randn(N,1)*sigma;
    noisy_signal = fo + noise;

    % Compute distance-based Laplacian 
    d = distanz([XCoords,YCoords]'); 
    A = exp(-d.^2/(2*s^2)); 
    A(A<thresh) = 0; % Thresholding to have a sparse adjacency matrix
    L = sgwt_laplacian(A);

    % Check graph connectivity
    connecteds(iter)=sgwt_check_connected(L);
    
    if connecteds(iter)
        % Solve the regularization problem:
        % min_f f'Lf + tau/2 * norm(f-fo)_2^2
        % by applying the filter tau/(tau+2lambda_l)
        
        % Compute Chebyshev coefficients of the multiplier
        lmax=sgwt_rough_lmax(L);
        arange=[0,lmax];
        cheb_coeffs=sgwt_cheby_coeff(smoothing_filter,K,K+1,arange);

        % Perform the approximate filtering
        smoothed_signal=sgwt_cheby_op(noisy_signal,L,cheb_coeffs,arange);

        % Solve the exact and approximate LASSO problems, equations (17)
        % and (19), respectively
        
        % Design spectral graph wavelet kernels
        [g,gp,t]=sgwt_filter_design(lmax,num_scales); 
        c=cell(numel(g),1);
        for i=1:numel(g)
            c{i}=sgwt_cheby_coeff(g{i},K,K+1,arange);
        end

        % Spectral Graph Wavelet Exact Operators
        Analysis_Exact=@(s) exact_forward2(s,g,L); 
        Synthesis_Exact=@(s) sgwt_adjoint_exact2(s,L,g);
        
        % Spectral Graph Wavelet Approximate Operators
        Analysis_Op=@(s) reshape(cell2mat(sgwt_cheby_op(s,L,c,arange)),N*(num_scales+1),1);
        Synthesis_Op=@(s) sgwt_adjoint2(s,L,c,arange);
        
        % Solve the exact and approximate LASSO problems, equations (17)
        % and (19), respectively
        [WISTA_exact_reconstruction, WISTA_exact_reconstructed_coeffs]=WISTA2(parameters,weights,noisy_signal,num_wavelets,Synthesis_Exact,Analysis_Exact);
        [WISTA_approx_reconstruction, WISTA_approx_reconstructed_coeffs]=WISTA2(parameters,weights,noisy_signal,num_wavelets,Synthesis_Op,Analysis_Op);
       
        % Compute MSEs of all 3 methods and the original noisy signal
        noise_MSEs(iter) = mean((noisy_signal-fo).^2);
        reconstruction_MSEs(iter)= mean((smoothed_signal-fo).^2);
        reconstruction_approx_wavelet_MSEs(iter)= mean((WISTA_approx_reconstruction-fo).^2);
        reconstruction_exact_wavelet_MSEs(iter)= mean((WISTA_exact_reconstruction-fo).^2);
    end
end

% Output of the numerical experiments
fprintf('%d out of %d graphs are connected.\n',sum(connecteds),num_iterations)
fprintf('The mean MSE of noisy signals on connected graphs is %f.\n',mean(noise_MSEs(connecteds==1)))
fprintf('The mean MSE of Tikhonov regularization reconstructed signals on connected graphs is %f.\n',mean(reconstruction_MSEs(connecteds==1)))
fprintf('The mean MSE of approximate weighted lasso reconstructed signals on connected graphs is %f.\n',mean(reconstruction_approx_wavelet_MSEs(connecteds==1)))
fprintf('The mean MSE of exact weighted lasso reconstructed signals on connected graphs is %f.\n',mean(reconstruction_exact_wavelet_MSEs(connecteds==1)))
