%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%											
%   "Copyright (c) 2011 David I Shuman"				
%   All rights reserved.
%
%   This is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This code is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   If you do not have a copy of the GNU General Public License,
%   see <http://www.gnu.org/licenses/>.									
%											
%   Author: David I Shuman
%   Signal Processing Laboratories LTS2 and LTS4
%   cole Polytechnique Fdrale de Lausanne (EPFL)
%   Contact: david.shuman@epfl.ch
%
%   Date last modified:	11/14/2011 
% 
%   This file requires the Spectral Graph Wavelets (SGWT) Toolbox, which is
%   is written by David K. Hammond, and is available at
%   http://wiki.epfl.ch/sgwt
%
%   Description:
%     This file executes the numerical comparison between the distributed 
%     Chebyshev approximation and two variations of the Jacobi method from 
%     Section VII.C of the paper "Distributed Signal Processing via 
%     Chebyshev Polynomial Approximation," which has been submitted to the 
%     IEEE Transactions on Signal Processing. Figure 6 of the paper is 
%     generated from this code. In order to generate the 4 subplots, we ran
%     this code 4 times with 4 different choices of P. Note that the code 
%     implements the computations in a centralized manner.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
close all;
randn('seed', 18); rand('seed', 18);

% Parameters of the random graph from Figure 2
N = 500;
plane_dim = 1;
sigma=.5; 
target_dist_cutoff = .075;
thresh = .6; 
s = sqrt(-target_dist_cutoff^2/(2*log(thresh)));
distance_cutoff = sqrt(-2*s^2*log(thresh));
tau=.5; 
smoothing_filter = @(x) tau./(tau+x);

% Generate data on a plane
XCoords = plane_dim*rand(N,1);
YCoords = plane_dim*rand(N,1);

% Generate the true signal
fo = 20*rand(N,1)-10; 

% Compute distance-based Laplacian 
d = distanz([XCoords,YCoords]'); 
W = exp(-d.^2/(2*s^2)); 
W(W<thresh) = 0; % Thresholding to have a sparse adjacency matrix
W=W-diag(diag(W)); % Set self-weights to 0
L = sgwt_laplacian(W);
D = diag(sum(W));
L_norm=D^(-.5)*L*D^(-.5);

% Choose a P matrix. Each subplot of Figure 6 is created from a different
% choice of P, corresponding to different centralized transductive
% classification methods
P_choice = 'random_walk';
switch P_choice
    case 'comb'
        r=2;
        P=L^r;
    case 'normalized'
        r=1;
        P=L_norm^r;
    case 'label_var'
        P=L*D^(-1);
    case 'random_walk'
        sigma=2;
        r=3;
        P=(sigma*eye(N)-L_norm)^(-r);
end

% Define y by filtering by 1/g, where g=tau/(tau+lambda)
y=(1/tau)*(tau*eye(N)+P)*fo;

% Now try different distributed iterative methods to recover f0 from y and 
% P by solving system of linear equations: (tau*I+P)*f=tau*y

% Set range of approximation/iteration orders (i.e. the amount of allowed
% communication)
minK=5;
maxK=60;
numK=maxK-minK+1;
Ks=minK:maxK;
errors_Cheb=zeros(1,numK);
errors_Jacobi=zeros(1,numK);
errors_Jacobi_Cheb=zeros(1,numK);

% Our distributed Chebyshev approximation
lmax=sgwt_rough_lmax(P);
arange=[0,lmax];
for K=minK:maxK
    % Compute Chebyshev coefficients of the multiplier
    cheb_coeffs=sgwt_cheby_coeff(smoothing_filter,K,K+1,arange);

    % Perform the approximate filtering
    smoothed_signal =sgwt_cheby_op(y,P,cheb_coeffs,arange);
    errors_Cheb(K-minK+1)=norm(fo-smoothed_signal);
end

% The Jacobi iterative method
P_D=diag(diag(P));
P_O=P-P_D;
iMat=(tau*eye(N)+P_D)^(-1);
c=iMat*tau*y;
R=-iMat*P_O;
spectral_radius_R=max(abs(eig(R)));
num_Jac_its=maxK; 
f_Jacobi=zeros(N,num_Jac_its+1); 
f_Jacobi(:,1)=y;
for i=2:num_Jac_its+1
    f_Jacobi(:,i)=R*f_Jacobi(:,i-1)+c;
    if (i-1) >= minK
        errors_Jacobi(i-minK)=norm(fo-f_Jacobi(:,i));
    end
end

% The Jacobi iterative method with Chebyshev acceleration 
% See for example pg. 298 of J. W. Demmel, Applied Numerical Linear Algebra,
% SIAM, 1997
rho=spectral_radius_R;
f_Jacobi_Cheb=zeros(N,num_Jac_its);
f_Jacobi_Cheb(:,1)=f_Jacobi(:,1);
f_Jacobi_Cheb(:,2)=f_Jacobi(:,2);
mu_two_prior=1;
mu_prior=rho;
for i=3:num_Jac_its+1
    mu_current=1/(2/(rho*mu_prior)-1/mu_two_prior);
    f_Jacobi_Cheb(:,i)=2*mu_current/(rho*mu_prior)*R*f_Jacobi_Cheb(:,i-1)-(mu_current/mu_two_prior)*f_Jacobi_Cheb(:,i-2)+(2*mu_current/(rho*mu_prior))*c;
    mu_two_prior=mu_prior;
    mu_prior=mu_current;
    if (i-1) >= minK
        errors_Jacobi_Cheb(i-minK)=norm(fo-f_Jacobi_Cheb(:,i));
    end
end

% Plot error curves
figure;
semilogy(Ks,errors_Cheb,'r',Ks,errors_Jacobi,'b-.',Ks,errors_Jacobi_Cheb,'k--','LineWidth',2);
h_legend=LEGEND('Our distributed Chebyshev approximation method', 'Jacobi method', 'Jacobi method with Chebyshev acceleration','Location','NorthWest'); 
set(h_legend,'FontSize',12);
LEGEND BOXOFF;
xlabel('Iterations / Approximation Order','FontSize',14);
ylabel('Error','FontSize',14);
set(gca,'YMinorTick','off')
set(gca,'FontSize',14);
set(gca,'XLim',[minK,maxK]);